
/*
 * error.c
 *   The functions in this file represent the libpqtypes error
 *   system.  It offers the API user the ability to set/get errors.
 *   The error system uses a per-thread global error, implemented
 *   using TLS (via declspec(thread) or pthread TLS).  For
 *   non-PQT_THREAD_SAFE builds, a static buffer is used.
 *
 * Copyright (c) 2008 eSilo, LLC. All rights reserved.
 * This is free software; see the source for copying conditions.  There is
 * NO warranty; not even for MERCHANTABILITY or  FITNESS FOR A  PARTICULAR
 * PURPOSE.
 */

#include "libpqtypes-int.h"

typedef struct
{
	int size;
	char *msg;
} pqterr_t;

/* non thread-safe, use a static */
#ifndef PQT_THREAD_SAFE
static pqterr_t lasterr = {0};

/* MSVC always uses __declspec(thread) */
#elif defined(PQT_MSVC)
static __declspec(thread) pqterr_t lasterr = {0};

/* Unix-variant, MinGW or Cygwin with no TLS storage class. */
#else
#include <pthread.h>
static pthread_key_t tlskey_lasterr;

/* When the thread dies, this will get called. */
static void tls_free_lasterr(void *value)
{
	if(value)
	{
		pqterr_t *e = (pqterr_t *)value;
		if(e->msg)
			free(e->msg);
		free(value);
		pthread_setspecific(tlskey_lasterr, NULL);
	}
}

/* called before main.  This attribute is available since gcc 2.7.0.
 * It replaced the obsolete _init ... destructor replaced _fini.
 * We used to use pthread_once, but solaris doesn't support this
 * in older versions (its simply a stub returning 0, ouch!).
 */
void __attribute__ ((constructor)) my_init(void)
{
	pthread_key_create(&tlskey_lasterr, tls_free_lasterr);
}
#endif

static void
vseterror(const char *format, va_list ap, int append);

static pqterr_t *
geterr(void)
{
	pqterr_t *err = NULL;

	/* Non thread-safe mode or windows msvc. */
#if !defined(PQT_THREAD_SAFE) || defined(PQT_MSVC)
	err = &lasterr;

	/* systems requiring pthread TLS keys */
#else

	/* System is using FSU Threads, like SCO OpenServer 5, which
	 * has minor prototype differences.
	 */
#	ifdef PTHREAD_FSU
		pthread_getspecific(tlskey_lasterr, (void **)&err);
#	else
		err = (pqterr_t *)pthread_getspecific(tlskey_lasterr);
#	endif

	if(!err)
	{
		err = (pqterr_t *)malloc(sizeof(pqterr_t));
		if(!err)
			return NULL;
		memset(err, 0, sizeof(pqterr_t));
		pthread_setspecific(tlskey_lasterr, err);
	}
#endif

	if(err->size == 0)
	{
		err->size = 256;
		err->msg = (char *)malloc(err->size);
		*err->msg = 0;
	}

	return err;
}

const char *
PQgeterror(void)
{
	pqterr_t *err = geterr();
	return (const char *)(err ? err->msg : NULL);
}

void
PQseterror(const char *format, ...)
{
	/* clear error message by passing in NULL, PQseterror(NULL).*/
	if(!format)
	{
		pqterr_t *err = geterr();
		if(err)
			*err->msg = 0;
	}
	else
	{
		va_list ap;
		va_start(ap, format);
		vseterror(format, ap, FALSE);
	}
}

/* errorf() callback for PGtypeArgs, see PQputf() and PQgetf().
 * Always returns -1.
 */
int
pqt_argserrorf(PGtypeArgs *args, const char *format, ...)
{
	va_list ap;
	char fqtn[200];

	if(!args || !format || !*format)
		return -1;

	pqt_fqtn(fqtn, sizeof(fqtn), args->typhandler->typschema,
		args->typhandler->typname);

	/* put the header */
	PQseterror("%s[pos:%d] - ", fqtn, args->typpos);

	/* append message from type handler */
	va_start(ap, format);
	vseterror(format, ap, TRUE);
	return -1;
}

static void
vseterror(const char *format, va_list ap, int append)
{
	int n;
	int curlen=0;
	char *p;
	va_list ap2;
	pqterr_t *err = geterr();

	if(!err)
		return;

	if(append)
		curlen = (int)strlen(err->msg);
	else
		*err->msg = 0;

  while(1)
	{
		va_copy(ap2, ap);
		n = pqt_vsnprintf(err->msg+curlen, err->size-curlen, format, ap2);
		va_end(ap2);

		/* success */
		if(n > -1)
			break;

		/* need more space */
		n = err->size * 2;
		if(!(p = pqt_realloc(err->msg, n)))
		{
			*err->msg = 0;
			break;
		}

		err->msg = p;
		err->size = n;
	}
}

